#
# database.rb
#
#   Copyright (c) 2001 Minero Aoki <aamine@loveruby.net>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

require 'refe/mindex'
require 'refe/findex'
require 'gdbm'


module ReFe

  class SearchError < StandardError; end


  class Database

    begin
      require 'refe/config'
    rescue LoadError
      DATA_DIR = '.'
    end
  
    def self.default_dir
      DATA_DIR + '/refe'
    end

    def self.load( dir = nil, readonly = true, init = false )
      new( dir || default_dir, readonly, init )
    end

    def save
      [ ['mindex', @mindex],
        ['findex', @findex],
        ['mtof', @mtof] ].each do |name, obj|
        File.open( @dir + '/' + name, 'w' ) {|f|
          Marshal.dump obj, f
        }
      end
    end


    def initialize( dir, ronly, init )
      @dir = dir
      @readonly = ronly
      @init = @readonly ? false : init

      @mindex = nil
      @mtof = nil
      @findex = nil
    end

    def method_index
      unless @mindex then
        @mindex = loadfromfile('mindex', MethodIndex)
        @mindex.descriptions = loadindex(@dir + '/mdesc')
      end
      @mindex
    end

    def method_to_function
      @mtof ||= loadfromfile('mtof', Hash)
    end

    def function_index
      unless @findex then
        @findex = loadfromfile('findex', FunctionIndex)
        @findex.descriptions = loadindex(@dir + '/fdesc')
        @findex.definitions  = loadindex(@dir + '/fdefs')
      end
      @findex
    end

    def loadfromfile( fname, klass )
      return klass.new if @init

      path = @dir + '/' + fname
      if File.exist? path then
        File.open( path ) {|f|
          return Marshal.load(f)
        }
      else
        @readonly and
            raise "refe database not initialized on #{@dir}: mkrefe first"
        klass.new
      end
    end

    def loadindex( fname )
      if @init then
        File.unlink fname if File.file? fname
      end
      Index.new( fname )
    end

    def method_index=( mi )
      @mindex = mi
    end

    def method_to_function=( mtof )
      @mtof = mtof
    end

    def function_index=( fi )
      @findex = fi
    end

  end


  class Index

    def initialize( fname )
      @filename = fname
      @conn = nil
      @class = GDBM
    end

    def updating
      if @conn then
        yield
        return
      end
      @class.open( @filename ) {|c|
        @conn = c
        yield
        @conn = nil
      }
    end

    def []( key )
      if @conn then
        @conn[key]
      else
        @class.open(@filename) {|f|
          return f[key]
        }
      end
    end

    def []=( key, val )
      @conn or raise IOError, 'call #updating before []='
      @conn[key] = val
    end

    def size
      if @conn then
        @conn.size
      else
        @class.open( @filename ) {|f|
          return f.size
        }
      end
    end

  end

end
